#!/usr/bin/env python3
import basf2
from basf2 import pickle_path as b2pp
path = b2pp.get_path_from_file("{{ pickle_file_path }}")
basf2.print_path(path)
basf2.process(path, max_event={{ max_event }})
print(basf2.statistics)
